@@ -38,6 +38,7 @@ import (
 
 	"github.com/prometheus/prometheus/config"
 	"github.com/prometheus/prometheus/discovery/targetgroup"
+	"github.com/prometheus/prometheus/pkg/exemplar"
 	"github.com/prometheus/prometheus/pkg/labels"
 	"github.com/prometheus/prometheus/pkg/pool"
 	"github.com/prometheus/prometheus/pkg/relabel"
@@ -142,19 +143,19 @@ var (
 	targetScrapeSampleDuplicate = prometheus.NewCounter(
 		prometheus.CounterOpts{
 			Name: "prometheus_target_scrapes_sample_duplicate_timestamp_total",
-			Help: "Total number of samples rejected due to duplicate timestamps but different values",
+			Help: "Total number of samples rejected due to duplicate timestamps but different values.",
 		},
 	)
 	targetScrapeSampleOutOfOrder = prometheus.NewCounter(
 		prometheus.CounterOpts{
 			Name: "prometheus_target_scrapes_sample_out_of_order_total",
-			Help: "Total number of samples rejected due to not being out of the expected order",
+			Help: "Total number of samples rejected due to not being out of the expected order.",
 		},
 	)
 	targetScrapeSampleOutOfBounds = prometheus.NewCounter(
 		prometheus.CounterOpts{
 			Name: "prometheus_target_scrapes_sample_out_of_bounds_total",
-			Help: "Total number of samples rejected due to timestamp falling outside of the time bounds",
+			Help: "Total number of samples rejected due to timestamp falling outside of the time bounds.",
 		},
 	)
 	targetScrapeCacheFlushForced = prometheus.NewCounter(
@@ -163,6 +164,12 @@ var (
 			Help: "How many times a scrape cache was flushed due to getting big while scrapes are failing.",
 		},
 	)
+	targetScrapeExemplarOutOfOrder = prometheus.NewCounter(
+		prometheus.CounterOpts{
+			Name: "prometheus_target_scrapes_exemplar_out_of_order_total",
+			Help: "Total number of exemplar rejected due to not being out of the expected order.",
+		},
+	)
 )
 
 func init() {
@@ -184,6 +191,7 @@ func init() {
 		targetScrapePoolTargetsAdded,
 		targetScrapeCacheFlushForced,
 		targetMetadataCache,
+		targetScrapeExemplarOutOfOrder,
 	)
 }
 
@@ -1245,9 +1253,10 @@ func (sl *scrapeLoop) getCache() *scrapeCache {
 }
 
 type appendErrors struct {
-	numOutOfOrder  int
-	numDuplicates  int
-	numOutOfBounds int
+	numOutOfOrder         int
+	numDuplicates         int
+	numOutOfBounds        int
+	numExemplarOutOfOrder int
 }
 
 func (sl *scrapeLoop) append(app storage.Appender, b []byte, contentType string, ts time.Time) (total, added, seriesAdded int, err error) {
@@ -1272,6 +1281,7 @@ loop:
 		var (
 			et          textparse.Entry
 			sampleAdded bool
+			e           exemplar.Exemplar
 		)
 		if et, err = p.Next(); err != nil {
 			if err == io.EOF {
@@ -1308,20 +1318,19 @@ loop:
 			continue
 		}
 		ce, ok := sl.cache.get(yoloString(met))
+		var (
+			ref  uint64
+			lset labels.Labels
+			mets string
+			hash uint64
+		)
 
 		if ok {
-			err = app.AddFast(ce.ref, t, v)
-			_, err = sl.checkAddError(ce, met, tp, err, &sampleLimitErr, &appErrs)
-			// In theory this should never happen.
-			if err == storage.ErrNotFound {
-				ok = false
-			}
-		}
-		if !ok {
-			var lset labels.Labels
-
-			mets := p.Metric(&lset)
-			hash := lset.Hash()
+			ref = ce.ref
+			lset = ce.lset
+		} else {
+			mets = p.Metric(&lset)
+			hash = lset.Hash()
 
 			// Hash label set as it is seen local to the target. Then add target labels
 			// and relabeling and store the final label set.
@@ -1337,17 +1346,18 @@ loop:
 				err = errNameLabelMandatory
 				break loop
 			}
+		}
 
-			var ref uint64
-			ref, err = app.Add(lset, t, v)
-			sampleAdded, err = sl.checkAddError(nil, met, tp, err, &sampleLimitErr, &appErrs)
-			if err != nil {
-				if err != storage.ErrNotFound {
-					level.Debug(sl.l).Log("msg", "Unexpected error", "series", string(met), "err", err)
-				}
-				break loop
+		ref, err = app.Append(ref, lset, t, v)
+		sampleAdded, err = sl.checkAddError(ce, met, tp, err, &sampleLimitErr, &appErrs)
+		if err != nil {
+			if err != storage.ErrNotFound {
+				level.Debug(sl.l).Log("msg", "Unexpected error", "series", string(met), "err", err)
 			}
+			break loop
+		}
 
+		if !ok {
 			if tp == nil {
 				// Bypass staleness logic if there is an explicit timestamp.
 				sl.cache.trackStaleness(hash, lset)
@@ -1362,6 +1372,18 @@ loop:
 		// number of samples remaining after relabeling.
 		added++
 
+		if hasExemplar := p.Exemplar(&e); hasExemplar {
+			if !e.HasTs {
+				e.Ts = t
+			}
+			_, exemplarErr := app.AppendExemplar(ref, lset, e)
+			exemplarErr = sl.checkAddExemplarError(exemplarErr, e, &appErrs)
+			if exemplarErr != nil {
+				// Since exemplar storage is still experimental, we don't fail the scrape on ingestion errors.
+				level.Debug(sl.l).Log("msg", "Error while adding exemplar in AddExemplar", "exemplar", fmt.Sprintf("%+v", e), "err", exemplarErr)
+			}
+		}
+
 	}
 	if sampleLimitErr != nil {
 		if err == nil {
@@ -1379,10 +1401,13 @@ loop:
 	if appErrs.numOutOfBounds > 0 {
 		level.Warn(sl.l).Log("msg", "Error on ingesting samples that are too old or are too far into the future", "num_dropped", appErrs.numOutOfBounds)
 	}
+	if appErrs.numExemplarOutOfOrder > 0 {
+		level.Warn(sl.l).Log("msg", "Error on ingesting out-of-order exemplars", "num_dropped", appErrs.numExemplarOutOfOrder)
+	}
 	if err == nil {
 		sl.cache.forEachStale(func(lset labels.Labels) bool {
 			// Series no longer exposed, mark it stale.
-			_, err = app.Add(lset, defTime, math.Float64frombits(value.StaleNaN))
+			_, err = app.Append(0, lset, defTime, math.Float64frombits(value.StaleNaN))
 			switch errors.Cause(err) {
 			case storage.ErrOutOfOrderSample, storage.ErrDuplicateSampleForTimestamp:
 				// Do not count these in logging, as this is expected if a target
@@ -1436,6 +1461,20 @@ func (sl *scrapeLoop) checkAddError(ce *cacheEntry, met []byte, tp *int64, err e
 	}
 }
 
+func (sl *scrapeLoop) checkAddExemplarError(err error, e exemplar.Exemplar, appErrs *appendErrors) error {
+	switch errors.Cause(err) {
+	case storage.ErrNotFound:
+		return storage.ErrNotFound
+	case storage.ErrOutOfOrderExemplar:
+		appErrs.numExemplarOutOfOrder++
+		level.Debug(sl.l).Log("msg", "Out of order exemplar", "exemplar", fmt.Sprintf("%+v", e))
+		targetScrapeExemplarOutOfOrder.Inc()
+		return nil
+	default:
+		return err
+	}
+}
+
 // The constants are suffixed with the invalid \xff unicode rune to avoid collisions
 // with scraped metrics in the cache.
 const (
@@ -1499,37 +1538,31 @@ func (sl *scrapeLoop) reportStale(app storage.Appender, start time.Time) (err er
 
 func (sl *scrapeLoop) addReportSample(app storage.Appender, s string, t int64, v float64) error {
 	ce, ok := sl.cache.get(s)
+	var ref uint64
+	var lset labels.Labels
 	if ok {
-		err := app.AddFast(ce.ref, t, v)
-		switch errors.Cause(err) {
-		case nil:
-			return nil
-		case storage.ErrNotFound:
-			// Try an Add.
-		case storage.ErrOutOfOrderSample, storage.ErrDuplicateSampleForTimestamp:
-			// Do not log here, as this is expected if a target goes away and comes back
-			// again with a new scrape loop.
-			return nil
-		default:
-			return err
+		ref = ce.ref
+		lset = ce.lset
+	} else {
+		lset = labels.Labels{
+			// The constants are suffixed with the invalid \xff unicode rune to avoid collisions
+			// with scraped metrics in the cache.
+			// We have to drop it when building the actual metric.
+			labels.Label{Name: labels.MetricName, Value: s[:len(s)-1]},
 		}
-	}
-	lset := labels.Labels{
-		// The constants are suffixed with the invalid \xff unicode rune to avoid collisions
-		// with scraped metrics in the cache.
-		// We have to drop it when building the actual metric.
-		labels.Label{Name: labels.MetricName, Value: s[:len(s)-1]},
+		lset = sl.reportSampleMutator(lset)
 	}
 
-	hash := lset.Hash()
-	lset = sl.reportSampleMutator(lset)
-
-	ref, err := app.Add(lset, t, v)
+	ref, err := app.Append(ref, lset, t, v)
 	switch errors.Cause(err) {
 	case nil:
-		sl.cache.addRef(s, ref, lset, hash)
+		if !ok {
+			sl.cache.addRef(s, ref, lset, lset.Hash())
+		}
 		return nil
 	case storage.ErrOutOfOrderSample, storage.ErrDuplicateSampleForTimestamp:
+		// Do not log here, as this is expected if a target goes away and comes back
+		// again with a new scrape loop.
 		return nil
 	default:
 		return err
